import { spawnSync } from 'child_process';
import { unlinkSync, readFileSync } from 'fs';
import { EventEmitter } from 'events';

export enum VideoType {
  YOUTUBE = 'youtube',
}

interface ClipVideoBaseParams {
  url: string;
  start?: number | string;
  end?: number | string;
  postprocessorExecutablePath?: string;
}

interface ClipYouTubeVideoParams extends ClipVideoBaseParams {
  downloaderExecutablePath?: string;
}

interface ClipVideoParams extends ClipYouTubeVideoParams {
  type: VideoType,
}

const convertSecondsToDuration = (s: number) => {
  const milliseconds = (s - Math.floor(s)) * 1000;
  const seconds = s % 60;
  const minutes = Math.floor(s / 60) % 60;
  const hours = Math.floor(s / 3600);
  const sss = milliseconds.toString().padStart(3, '0');
  const ss = seconds.toString().padStart(2, '0');
  const mm = minutes.toString().padStart(2, '0');
  const hh = hours.toString().padStart(2, '0');
  return `${hh}:${mm}:${ss}.${sss}`;
};

const convertDurationToSeconds = (d: string) => {
  const [hh, mm, ss] = d.split(':');
  return (Number(hh) * 3600) + (Number(mm) * 60) + Number(ss);
};

const normalizeStartSeconds = (start?: number | string) => {
  if (typeof start === 'undefined') {
    return 0;
  }
  if (typeof start === 'string') {
    return convertDurationToSeconds(start);
  }

  return start;
};

const getCacheFilename = () => {
  const time = new Date().getTime();
  return `output-${time}.mkv`;
};

export interface ClipEventEmitter extends EventEmitter {
  process(): void;
}

class YouTubeVideoClipEventEmitter extends EventEmitter implements ClipEventEmitter {
  constructor(private readonly params: ClipYouTubeVideoParams) {
    super();
  }

  private constructDefaultDownloadArgs(cacheFilename: string) {
    const startSeconds = normalizeStartSeconds(this.params.start);

    const downloadArgs = [
      '--youtube-skip-dash-manifest',
      '-f', 'bestvideo+bestaudio',
      '--fragment-retries', 'infinite',
    ];

    if (typeof this.params.end !== 'undefined') {
      downloadArgs.push('--download-sections');
      const endSeconds = (
        typeof this.params.end === 'string'
          ? convertDurationToSeconds(this.params.end)
          : this.params.end
      );
      downloadArgs.push(
        `*${convertSecondsToDuration(startSeconds)}-${convertSecondsToDuration(endSeconds)}`,
      );
      downloadArgs.push('-o');
      downloadArgs.push(cacheFilename);
      downloadArgs.push('--force-keyframes-at-cuts');
      if (startSeconds > 0) {
        downloadArgs.push(`${this.params.url}?t=${startSeconds}`);
      } else {
        downloadArgs.push(this.params.url);
      }
    } else if (startSeconds > 0) {
      downloadArgs.push('--download-sections');
      downloadArgs.push(`*${convertSecondsToDuration(startSeconds)}-inf`);
      downloadArgs.push('-o');
      downloadArgs.push(cacheFilename);
      downloadArgs.push('--force-keyframes-at-cuts');
      downloadArgs.push(`${this.params.url}?t=${startSeconds}`);
    } else {
      downloadArgs.push('-o');
      downloadArgs.push(cacheFilename);
      downloadArgs.push(this.params.url);
    }

    return downloadArgs;
  }

  private doDownloadProcess(downloadArgs: string[]) {
    this.emit('process', {
      type: 'download',
      phase: 'start',
      command: `${this.params.downloaderExecutablePath as string} ${downloadArgs.join(' ')}`,
    });
    const downloaderProcess = spawnSync(
      this.params.downloaderExecutablePath as string,
      downloadArgs,
    );
    if (downloaderProcess.error) {
      this.emit('error', new Error('Downloader error.', { cause: downloaderProcess.error }));
      this.emit('end', null);
      return null;
    }
    return downloaderProcess;
  }

  private processDefault() {
    const cacheFilename = getCacheFilename(); // todo label this on the cache
    const downloadArgs = this.constructDefaultDownloadArgs(cacheFilename);
    const downloaderProcess = this.doDownloadProcess(downloadArgs);
    if (!downloaderProcess) {
      return;
    }

    this.emit('process', {
      type: 'download',
      phase: 'success',
    });

    const output = readFileSync(cacheFilename);
    unlinkSync(cacheFilename);
    this.emit('success', {
      output,
    });
    this.emit('end');
  }

  private processWithPostprocessing() {
    const downloadArgs = [
      '--youtube-skip-dash-manifest',
      '-f', 'bestvideo+bestaudio',
      '-g', this.params.url,
    ];
    const downloaderProcess = this.doDownloadProcess(downloadArgs);
    if (!downloaderProcess) {
      return;
    }

    const [videoUrlStream, audioUrlStream] = downloaderProcess.stdout.toString('utf-8').split('\n');
    this.emit('process', {
      type: 'download',
      phase: 'success',
      streamUrls: {
        video: [videoUrlStream],
        audio: [audioUrlStream],
      },
    });

    const startSeconds = normalizeStartSeconds(this.params.start);
    const ffmpegProcessVideoStreamArgs = [
      '-ss', (this.params.start ?? '00:00:00').toString(),
      '-i', videoUrlStream,
    ];

    const ffmpegProcessAudioStreamArgs = [
      '-ss', (this.params.start ?? '00:00:00').toString(),
      '-i', audioUrlStream,
    ];

    if (typeof this.params.end !== 'undefined') {
      // -to flag is broken on video stream in ffmpeg.....
      const endSeconds = (
        typeof this.params.end === 'string'
          ? convertDurationToSeconds(this.params.end)
          : this.params.end
      );

      const difference = endSeconds - startSeconds;
      const clipDuration = convertSecondsToDuration(difference);

      ffmpegProcessVideoStreamArgs.push('-t');
      ffmpegProcessVideoStreamArgs.push(clipDuration);

      ffmpegProcessAudioStreamArgs.push('-t');
      ffmpegProcessAudioStreamArgs.push(clipDuration);
    }

    const cacheFilename = getCacheFilename(); // todo label this on the cache
    const ffmpegProcessArgs = [
      ...ffmpegProcessVideoStreamArgs,
      ...ffmpegProcessAudioStreamArgs,
      '-map', '0:v',
      '-map', '1:a',
      cacheFilename,
    ];
    this.emit('process', {
      type: 'postprocess',
      phase: 'start',
      command: `${this.params.postprocessorExecutablePath as string} ${ffmpegProcessArgs.join(' ')}`,
    });
    const ffmpegProcess = spawnSync(
      this.params.postprocessorExecutablePath as string,
      ffmpegProcessArgs,
    );
    if (ffmpegProcess.error) {
      this.emit('error', new Error('ffmpeg error.', { cause: ffmpegProcess.error }));
      this.emit('end');
      return;
    }
    this.emit('process', {
      type: 'postprocess',
      phase: 'success',
    });
    const output = readFileSync(cacheFilename);
    unlinkSync(cacheFilename);
    this.emit('success', {
      output,
    });
    this.emit('end');
  }

  process() {
    if (!this.params.downloaderExecutablePath) {
      this.emit('error', new Error('Downloader not found.'));
      this.emit('end', null);
      return;
    }

    if (this.params.postprocessorExecutablePath) {
      this.processWithPostprocessing();
      return;
    }
    this.processDefault();
  }
}

export const createVideoClipper = (params: ClipVideoParams) => {
  const {
    type: videoType,
    url,
    start,
    end,
    downloaderExecutablePath,
    postprocessorExecutablePath,
  } = params;

  switch (videoType as string) {
    case VideoType.YOUTUBE:
      return new YouTubeVideoClipEventEmitter({
        downloaderExecutablePath,
        postprocessorExecutablePath,
        url,
        start,
        end,
      });
    default:
      break;
  }

  throw new TypeError(`Invalid video type: "${videoType}". Valid values are: ${JSON.stringify(Object.values(VideoType))}`);
};
